'use strict';

var _createClass = function () { function defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } } return function (Constructor, protoProps, staticProps) { if (protoProps) defineProperties(Constructor.prototype, protoProps); if (staticProps) defineProperties(Constructor, staticProps); return Constructor; }; }();

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

/**
 * Classic wave path finding algorithm
 * Wiki: (https://en.wikipedia.org/wiki/Lee_algorithm)
 *
 * @author Andrey Zharikov & Nikolay Govorov
 *
 */

var WavePathfinder = function () {
  _createClass(WavePathfinder, null, [{
    key: 'findPath',

    /**
     * Shorthand method for findPath
     *
     * @param {array} obstaclesMatrix
     * @param {number} startX
     * @param {number} startY
     * @param {number} finishX
     * @param {number} finishY
     *
     * @return {null|object[]} Path from start to finish or null if not found.
     *                         Example: [ [0, 0], [1, 0], [2, 0] ]
     */

    value: function findPath(obstaclesMatrix, startX, startY, finishX, finishY) {
      var finder = new this(obstaclesMatrix);

      return finder.findPath(startX, startY, finishX, finishY);
    }

    /**
     * Instantiate object based on passability matrix
     *
     * @param {array} obstaclesMatrix - two-dimensional boolean array (x, y)
     *                where true is passable cell and false is non-passable.
     *                other truthy and falsy values can be used as well.
     *                Example: [ [1, 1, 1], [0, 0, 1], [1, 0, 1] ]
     *
     * @return {object} WavePathfinder object
     */

  }]);

  function WavePathfinder(obstaclesMatrix) {
    _classCallCheck(this, WavePathfinder);

    if (!Array.isArray(obstaclesMatrix)) {
      throw new Error('Incorrect matrix!');
    }

    this.START_CELL = 0;
    this.UNVISITED_CELL = -1;

    this.obstaclesMatrix = obstaclesMatrix.map(function (row) {
      return row.slice();
    });

    return this;
  }

  /**
   * Shorthand that combines expandWave + backtracePath.
   *
   * @param {number} startX
   * @param {number} startY
   * @param {number} finishX
   * @param {number} finishY
   *
   * @return {null|array} Shortest path from start to finish or null if not found.
   *                      Example: [ [0, 0], [1, 0], [2, 0] ]
   */

  _createClass(WavePathfinder, [{
    key: 'findPath',
    value: function findPath(startX, startY, finishX, finishY) {
      this.expandWave(startX, startY);

      return this.backtracePath(finishX, finishY);
    }

    /**
     * Expand wave and calculate steps number for each reachable cell.
     * This method must be called at least once to find shortest paths from start to other cells.
     * Once it has been called, backtracePath can be called multiple times for different finish cells.
     *
     * @param {number} startX
     * @param {number} startY
     *
     * @return {array} Steps array with minimum possible steps number for each cell.
     *                 0 means start (current) cell, -1 means unreachable cell.
     */

  }, {
    key: 'expandWave',
    value: function expandWave(startX, startY) {
      var _this = this;

      this.resultPath = [];

      // first part of the wave algorithm - work field matrix initialization
      this.stepsMatrix = this.obstaclesMatrix.map(function (row) {
        return row.slice().fill(_this.UNVISITED_CELL);
      });
      this.stepsMatrix[startX][startY] = this.START_CELL;

      // second part of the wave algorithm - wave propagation
      var propagateWave = function propagateWave(newX, newY, step) {
        if (_this.obstaclesMatrix[newX] && !_this.obstaclesMatrix[newX][newY]) {
          if (_this.stepsMatrix[newX][newY] === _this.UNVISITED_CELL) {
            _this.stepsMatrix[newX][newY] = step + 1;
          }
        }
      };

      for (var step = 0; step < this.stepsMatrix.length * this.stepsMatrix[0].length; step++) {
        for (var x = 0; x < this.stepsMatrix.length; x++) {
          for (var y = 0; y < this.stepsMatrix[0].length; y++) {
            if (this.stepsMatrix[x][y] === step) {
              propagateWave(x, y + 1, step); // up
              propagateWave(x + 1, y, step); // right
              propagateWave(x, y - 1, step); // down
              propagateWave(x - 1, y, step); // left
            }
          }
        }
      }

      return this.stepsMatrix;
    }

    /**
     * Finds shortest path from start to finish based on steps matrix.
     * Call this method after expandWave.
     *
     * @param {number} finishX
     * @param {number} finishY
     *
     * @return {null|array} Path from start to finish or null if not found.
     *                      Example: [ [0, 0], [1, 0], [2, 0] ]
     */

  }, {
    key: 'backtracePath',
    value: function backtracePath(finishX, finishY) {
      var _this2 = this;

      if (typeof this.stepsMatrix === 'undefined') {
        throw new Error('Call expandWave first to calculate stepsMatrix!');
      }

      if (this.stepsMatrix[finishX][finishY] === this.UNVISITED_CELL) {
        this.resultPath = null;
        return null;
      }

      this.resultPath = [];

      var currentX = finishX;
      var currentY = finishY;

      var addStep = function addStep(x, y) {
        _this2.resultPath.push([y, x]);
      };

      addStep(finishX, finishY);

      var propagateWave = function propagateWave(newX, newY, step) {
        if (_this2.stepsMatrix[newX] !== undefined) {
          if (_this2.stepsMatrix[newX][newY] === step - 1) {
            addStep(newX, newY);
            currentX = newX;
            currentY = newY;
            if (step === 1) return true;
          }
        }
        return false;
      };

      for (var step = this.stepsMatrix[finishX][finishY]; step >= 0; step--) {
        if (propagateWave(currentX + 1, currentY, step) || propagateWave(currentX - 1, currentY, step) || propagateWave(currentX, currentY + 1, step) || propagateWave(currentX, currentY - 1, step)) {
          break;
        }
      }

      return this.resultPath.reverse();
    }
  }]);

  return WavePathfinder;
}();

module.exports = WavePathfinder;